import re
from typing import overload

from LMS.Message.Definitions.Field.LMS_Field import LMS_Field
from LMS.Message.Tag.LMS_Tag import LMS_DecodedTag, LMS_EncodedTag, LMS_TagBase
from LMS.Message.Tag.System_Definitions import get_system_tag
from LMS.Message.Tag.Tag_Formats import (DECODED_FORMAT, ENCODED_FORMAT,
                                         TAG_FORMAT)
from LMS.TitleConfig.Config import TagConfig


class LMS_MessageText:
    """Class that represents a text entry."""

    def __init__(
        self, message: str | list[str | LMS_TagBase], config: TagConfig | None = None
    ):
        # Inject a reference to the tag definitions to allow for decoding of names and parameter types
        self._config = config

        if isinstance(message, str):
            self._set_parts(message)
        else:
            self._parts = message

    def __iter__(self):
        return iter(self._parts)

    @property
    def text(self) -> str:
        """The raw text of the message."""
        return "".join(
            part.to_text() if isinstance(part, LMS_TagBase) else part
            for part in self._parts
        )

    @text.setter
    def text(self, string: str) -> str:
        self._set_parts(string)

    @property
    def tags(self) -> list[LMS_EncodedTag | LMS_DecodedTag]:
        """The list of control tags in the message."""
        return [part for part in self._parts if isinstance(part, LMS_TagBase)]

    @overload
    def append_encoded_tag(
        self, group: int, tag: int, parameters: list[str] | None = None
    ) -> None: ...

    @overload
    def append_encoded_tag(
        self, group: str, tag: str, parameters: list[str] | None = None
    ) -> None: ...

    def append_encoded_tag(
        self, group: int | str, tag: int | str, parameters: list[str] | None = None
    ):
        """Appends an encoded tag to the current message.

        :param group: the group name or index.
        :param tag: the group tag or index:
        :param parameters: a list of hex strings.
        """
        if isinstance(group, int) and isinstance(tag, int):
            self._parts.append(LMS_EncodedTag(group, tag, parameters))
            return

        definition = self._config.get_definition_by_names(group, tag)
        self._parts.append(
            LMS_EncodedTag(
                definition.group_index, definition.tag_index, parameters, group, tag
            )
        )

    def append_decoded_tag(self, group_name: str, tag_name: str, **parameters) -> None:
        """Appends an decoded tag to the current message.

        :param group_name: the group name.
        :param tag_name: the tag name.:
        :param parameters: keyword arguments of parameters mapped to their value.
        """

        if group_name == "System":
            definition = get_system_tag(tag_name)
        else:
            definition = self._config.get_definition_by_names(group_name, tag_name)

        # The provided kwargs must match the structure defined in the config in order for value conversion to work
        converted_params = {
            param_def.name: LMS_Field(parameters[param_def.name], param_def)
            for param_def in definition.parameters
        }

        tag = LMS_DecodedTag(
            definition.group_index,
            definition.tag_index,
            group_name,
            tag_name,
            converted_params,
        )
        self._parts.append(tag)

    def append_tag_string(self, string: str) -> None:
        """Appends a tag to the current message given a string.

        :param string: the tag string."""
        if re.match(DECODED_FORMAT, string):
            tag = LMS_DecodedTag
        elif re.match(ENCODED_FORMAT, string):
            tag = LMS_EncodedTag
        else:
            raise ValueError(f"Invalid format for tag '{string}'.")

        self._parts.append(tag.from_string(string, self._config))

    def _set_parts(self, text: str) -> None:
        self._parts = []
        for part in re.split(TAG_FORMAT, text):
            if bool(re.match(TAG_FORMAT, part)):
                self.append_tag_string(part)
            else:
                self._parts.append(part)



import os
from typing import BinaryIO

from LMS.Common.LMS_DataType import LMS_DataType
from LMS.Common.Stream.FileInfo import read_file_info
from LMS.Common.Stream.Hashtable import read_labels
from LMS.Common.Stream.Section import read_section_data
from LMS.FileIO.Stream import FileReader
from LMS.Project.MSBP import MSBP
from LMS.Project.Section.ALI2 import read_ali2
from LMS.Project.Section.ATI2 import read_ati2
from LMS.Project.Section.CLR1 import read_clr1
from LMS.Project.Section.String import read_strings
from LMS.Project.Section.SYL3 import read_styles
from LMS.Project.Section.TAG2 import read_tag2
from LMS.Project.Section.TGG2 import read_tgg2
from LMS.Project.Section.TGP2 import read_tgp2


def read_msbp(stream: BinaryIO | None) -> MSBP:
    """Reads a MSBP file from a stream.

    :param stream: a stream object.

    ## Usage
    ```
    with open(file_path, "rb") as file:
        msbp = read_msbp(file)
        ...
    ```"""
    if stream is None:
        raise ValueError("Stream must be valid!")

    reader = FileReader(stream)
    file_info = read_file_info(reader, "MsgPrjBn")

    exist_map = {"CLB1": False, "ALB1": False, "SLB1": False}

    attribute_info_list = attribute_lists = tag_groups = tag_info_list = styles = (
        source_list
    ) = tag_param_list = None
    for magic, _ in read_section_data(reader, file_info.section_count):
        match magic:
            case "CLB1" | "ALB1" | "SLB1":
                labels, _ = read_labels(reader)
                # Set the name attributes of last read item
                for i in labels:
                    items[i].name = labels[i]
                exist_map[magic] = True
            case "CLR1":
                colors = read_clr1(reader)
                items = colors
            case "ATI2":
                attribute_info_list = read_ati2(reader)
                items = attribute_info_list
            case "ALI2":
                attribute_lists = read_ali2(reader)
            case "TGG2":
                tag_groups = read_tgg2(reader, file_info.version)
            case "TAG2":
                tag_info_list = read_tag2(reader)
            case "TGP2":
                tag_param_list = read_tgp2(reader)
            case "TGL2":
                list_items = read_strings(reader, 16)
            case "SYL3":
                styles = read_styles(reader)
                items = styles
            case "CTI1":
                source_list = read_strings(reader, 32)
            case _:
                pass

    # Set all list items for attributes
    if attribute_info_list:
        for info in attribute_info_list:
            if info.datatype is LMS_DataType.LIST:
                info.list_items = attribute_lists[info.list_index]

    # Combining tag data
    if tag_groups:

        # Cut out the System group from the definitions
        # Since these are predefined, there is no need to keep them.
        # Easier to remove here as opposed to manually skipping over them while reading
        # See Tag/System_Definition for the full System definition of tags.
        tag_groups = tag_groups[1:]

        for group in tag_groups:
            group.tag_definitions = [tag_info_list[i] for i in group.tag_indexes]

            for tag in group.tag_definitions:
                tag.param_info = [tag_param_list[i] for i in tag.parameter_indexes]

                for parameter in tag.param_info:
                    parameter.list_items = [
                        list_items[i] for i in parameter.list_indexes
                    ]

    file = MSBP(file_info, colors, attribute_info_list, tag_groups, styles, source_list)
    file.name = os.path.basename(stream.name).removesuffix(".msbp")
    return file



from __future__ import annotations

from dataclasses import dataclass, field

from LMS.Common.LMS_DataType import LMS_DataType
from LMS.TitleConfig.Definitions.Value import ValueDefinition


@dataclass(frozen=True)
class TagDefinition:
    """Class that represents a signle tag definition in the structure."""

    group_name: str
    group_index: int
    tag_name: str
    tag_index: int
    description: str
    parameters: list[ValueDefinition] = field(default=None)

    @classmethod
    def from_dict(cls, data: dict, group_map: dict[int, str]):
        tag_name = data["name"]
        group_index, tag_index = data["group_index"], data["tag_index"]
        group_name = group_map[group_index]
        tag_description = data["description"]

        group_name = group_map[group_index]

        parameters = None
        if "parameters" in data:
            parameters = []
            for param_def in data["parameters"]:
                param_name = param_def["name"]
                param_description = param_def["description"]
                datatype = LMS_DataType.from_string(param_def["datatype"])
                list_items = param_def.get("list_items")
                parameters.append(
                    ValueDefinition(param_name, param_description, datatype, list_items)
                )

        return cls(
            group_name,
            group_index,
            tag_name,
            tag_index,
            tag_description,
            parameters,
        )


@dataclass(frozen=True)
class TagConfig:
    """Class that represents an tag structure definition."""

    group_map: dict[int, str]
    definitions: list[TagDefinition]

    def get_definition_by_names(self, group: str, tag: str) -> TagDefinition:
        group_index = None
        for i, name in self.group_map.items():
            if name == group:
                group_index = i
                break

        if group_index is None:
            raise KeyError(f"Group name '{group}' was not found! Is the group defined?")

        for tag_def in self.definitions:
            if tag_def.group_index == group_index and tag_def.tag_name == tag:
                return tag_def

        raise KeyError(
            f"Tag name '{tag}' not found in group '{group}'. Is the tag defined?"
        )

    def get_definition_by_indexes(self, group: int, tag: int) -> TagDefinition:
        for tag_def in self.definitions:
            if tag_def.group_index == group and tag_def.tag_index == tag:
                return tag_def

        raise KeyError(
            f"Tag index of {tag} was not found in group index {group}. Is the tag defined?"
        )



from __future__ import annotations

import struct
from io import BytesIO
from typing import BinaryIO, Generator

from LMS.FileIO.Encoding import FileEncoding

STRUCT_TYPES = {
    "little": {
        "uint8": "<B",
        "uint16": "<H",
        "uint32": "<I",
        "int8": "<b",
        "int16": "<h",
        "int32": "<i",
        "float": "<f",
    },
    "big": {
        "uint8": ">B",
        "uint16": ">H",
        "uint32": ">I",
        "int8": ">b",
        "int16": ">h",
        "int32": ">i",
        "float": ">f",
    },
}


class FileReader:
    def __init__(self, stream: BinaryIO | bytes, big_endian: bool = False):
        if isinstance(stream, bytes):
            self._data = BytesIO(stream)
        else:
            self._data = BytesIO(stream.read())

        self.encoding: FileEncoding = None
        self.big_endian = big_endian

    def tell(self) -> int:
        return self._data.tell()

    def skip(self, length: int) -> None:
        self._data.read(length)

    def align(self, alignment: int) -> int:
        alignment = (-self.tell() % alignment + alignment) % alignment
        self.skip(alignment)

    def read_bytes(self, length: int) -> bytes:
        return self._data.read(length)

    def seek(self, offset: int, whence: int = 0) -> None:
        if offset < 0:
            self._data.seek(self.tell() + offset)
        else:
            self._data.seek(offset, whence)

    def read_offset_array(self, count: int) -> Generator[int, None, None]:
        start = self.tell() - 4
        for _ in range(count):
            last = self.tell() + 4
            yield self.read_uint32() + start
            self.seek(last)

    def read_uint16_array(self, count: int) -> list[int]:
        return [self.read_uint16() for _ in range(count)]

    def read_int8(self) -> int:
        return struct.unpack(self._get_datatype("int8"), self._data.read(1))[0]

    def read_int16(self) -> int:
        return struct.unpack(self._get_datatype("int16"), self._data.read(2))[0]

    def read_int32(self) -> int:
        return struct.unpack(self._get_datatype("int32"), self._data.read(4))[0]

    def read_uint8(self) -> int:
        return struct.unpack(self._get_datatype("uint8"), self._data.read(1))[0]

    def read_uint16(self) -> int:
        return struct.unpack(self._get_datatype("uint16"), self._data.read(2))[0]

    def read_uint32(self) -> int:
        return struct.unpack(self._get_datatype("uint32"), self._data.read(4))[0]

    def read_float32(self) -> float:
        return struct.unpack(self._get_datatype("float"), self._data.read(4))[0]

    def read_string_len(self, length: int) -> str:
        return self._data.read(length).decode("UTF-8")

    def read_str_variable_encoding(self):
        message = b""
        while (
            raw_char := self.read_bytes(self.encoding.width)
        ) != self.encoding.terminator:
            message += raw_char
        return message.decode(self.encoding.to_string_format(self.big_endian))

    def read_len_string_variable_encoding(self):
        self.align(len("\x00".encode(self.encoding.to_string_format(self.big_endian))))
        length = self.read_uint16()
        return self.read_bytes(length).decode(
            self.encoding.to_string_format(self.big_endian)
        )

    def _get_datatype(self, name: str) -> str:
        return STRUCT_TYPES["little" if not self.big_endian else "big"][name]


class FileWriter:
    def __init__(self, encoding: FileEncoding):
        self.data = BytesIO(b"")
        self.encoding = encoding
        self.big_endian = False

    def skip(self, length: int) -> None:
        self.data.seek(length, 1)

    def get_stream_size(self) -> int:
        last_position = self.tell()
        self.seek(0, 2)
        size = self.tell()
        self.seek(last_position)
        return size

    def write_bytes(self, data: bytes) -> bytes:
        return self.data.write(data)

    def seek(self, offset: int, whence: int = 0) -> None:
        self.data.seek(offset, whence)

    def tell(self) -> int:
        return self.data.tell()

    def write_alignment(self, data: bytes, alignment: int) -> None:
        self.write_bytes(data * self.align(self.tell(), alignment))

    def write_uint16_array(self, array: list[int]) -> None:
        for number in array:
            self.write_uint16(number)

    def write_int8(self, value: int) -> int:
        return self.data.write(struct.pack(self._get_datatype("int8"), value))

    def write_int16(self, value: int) -> int:
        return self.data.write(struct.pack(self._get_datatype("int16"), value))

    def write_int32(self, value: int) -> None:
        return self.data.write(struct.pack(self._get_datatype("int32"), value))

    def write_uint8(self, value: int) -> None:
        return self.data.write(struct.pack(self._get_datatype("uint8"), value))

    def write_uint16(self, value: int) -> None:
        return self.data.write(struct.pack(self._get_datatype("uint16"), value))

    def write_uint32(self, value: int) -> None:
        return self.data.write(struct.pack(self._get_datatype("uint32"), value))

    def write_float32(self, value: float) -> None:
        return self.data.write(struct.pack(self._get_datatype("float"), value))

    def write_string(self, string: str):
        self.write_bytes(string.encode("UTF-8"))

    def write_len_variable_encoding_string(self, string: str) -> None:
        self.write_uint16(len(string) * self.encoding.width)
        self.write_variable_encoding_string.write_variable_encoding_string(
            string, False
        )

    def write_variable_encoding_string(self, string: str, terminate: bool = True):
        self.write_bytes(string.encode(self.encoding.to_string_format(self.big_endian)))
        if terminate:
            self.write_bytes(b"\x00" * self.encoding.width)

    def align(self, number: int, alignment: int) -> int:
        return (-number % alignment + alignment) % alignment

    def _get_datatype(self, name: str) -> str:
        return STRUCT_TYPES["little" if not self.big_endian else "big"][name]

    def get_data(self) -> bytes:
        return self.data.getvalue()



from dataclasses import dataclass

from LMS.FileIO.Encoding import FileEncoding


@dataclass(frozen=True)
class LMS_FileInfo:
    big_endian: bool
    encoding: FileEncoding
    version: int
    section_count: int



from LMS.FileIO.Stream import FileReader


def read_strings(reader: FileReader, count_bits: int) -> list[str]:
    string_list = []

    match count_bits:
        case 16:
            count = reader.read_uint16()
            reader.skip(2)
        case 32:
            count = reader.read_uint32()

    for offset in reader.read_offset_array(count):
        reader.seek(offset)
        string_list.append(reader.read_str_variable_encoding())

    return string_list



from typing import Self

from LMS.Message.Definitions.Field.LMS_Field import LMS_Field
from LMS.Message.Definitions.Field.LMS_FieldMap import LMS_FieldMap
from LMS.Message.Definitions.LMS_MessageText import LMS_MessageText
from LMS.TitleConfig.Definitions.Attributes import AttributeConfig
from LMS.TitleConfig.Definitions.Tags import TagConfig


class MSBTEntry:
    """A class that represents an entry in a MSBT file."""

    def __init__(
        self,
        name: str,
        message: str | LMS_MessageText | None,
        attribute: LMS_FieldMap | bytes = None,
        style_index: int = None,
    ):
        self.name = name

        if isinstance(message, str):
            self._message = LMS_MessageText(message)
        else:
            self._message = message

        self._attribute = attribute
        self.style_index = style_index

    @property
    def message(self) -> LMS_MessageText:
        """The message object for the instance."""
        return self._message

    @property
    def attribute(self) -> LMS_FieldMap | bytes:
        """The attribute for the instance."""
        return self._attribute

    def to_dict(self) -> dict:
        """Converts the MSBTEntry instance into a dictionary object."""
        result = {"name": self.name}

        if self.message is not None:
            result["message"] = self.message.text

        if self.attribute is not None:
            if isinstance(self.attribute, bytes):
                result["attribute"] = self.attribute.hex().upper()
            else:
                result["attribute"] = {
                    name: attr.value for name, attr in self.attribute.items()
                }

        if self.style_index is not None:
            result["style_index"] = self.style_index

        return result

    @classmethod
    def from_dict(
        cls,
        data: dict,
        attribute_conifg: AttributeConfig = None,
        tag_config: TagConfig = None,
    ) -> Self:
        """Creates a MSBTEntry from a dictionary object.

        :param data: the dictionary data.
        :param attribute_config: the config to use to import decoded attributes.
        :param tag_config: the config to use if decoded tags are included in the message.
        """

        if "message" in data:
            message = LMS_MessageText(data["message"], tag_config)
        else:
            message = None

        if "attribute" in data:
            if isinstance(data["attribute"], dict):
                attribute = {
                    definition.name: LMS_Field(
                        data["attribute"][definition.name], definition
                    )
                    for definition in attribute_conifg.definitions
                }
            else:
                attribute = bytes.fromhex(data["attribute"])
        else:
            attribute = None

        if "style_index" in data:
            style_index = data["style_index"]
        else:
            style_index = None

        return cls(data["name"], message, attribute, style_index)



from LMS.Common.LMS_FileInfo import LMS_FileInfo
from LMS.Project.Definitions.Attribute import LMS_AttributeInfo
from LMS.Project.Definitions.Color import LMS_Color
from LMS.Project.Definitions.Style import LMS_Style
from LMS.Project.Definitions.Tag import LMS_TagGroup


class MSBP:
    """A class that represents a MSBP file.

    https://nintendo-formats.com/libs/lms/msbp.html."""

    def __init__(
        self,
        info: LMS_FileInfo,
        colors: list[LMS_Color],
        config: list[LMS_AttributeInfo],
        tag_groups: list[LMS_TagGroup],
        styles: list[LMS_Style],
        source_files: list[str],
    ):
        self.name = None

        self._info = info

        self._colors = colors
        self._config = config
        self._tag_groups = tag_groups
        self._styles = styles
        self._source_files = source_files

    @property
    def info(self) -> LMS_FileInfo:
        """The stream info of the MSBP."""
        return self._info

    @property
    def colors(self) -> list[LMS_Color]:
        """The color definitions for the project."""
        return self._colors

    @property
    def attribute_info(self) -> list[LMS_AttributeInfo]:
        """The attribute definitions for the project instance."""
        return self._config

    @property
    def tag_groups(self) -> list[LMS_TagGroup]:
        """The tag group definitions for the project instance."""
        return self._tag_groups

    @property
    def style_list(self) -> list[LMS_Style]:
        """The style definitions for the project instance."""
        return self._styles

    @property
    def source_files(self) -> list[str]:
        """The source file definitions for the project instance."""
        return self._source_files



class LMS_Style:
    def __init__(
        self,
        name: str,
        region_width: int = None,
        line_number: int = None,
        font_index: int = None,
        color_index: int = None,
    ):
        self.name = name
        self.region_width = region_width
        self.line_number = line_number
        self.font_index = font_index
        self.color_index = color_index



from LMS.FileIO.Stream import FileReader
from LMS.Project.Definitions.Tag import LMS_TagGroup


def read_tgg2(reader: FileReader, version: int) -> list[LMS_TagGroup]:
    group_list = []

    count = reader.read_uint32()
    for i, offset in enumerate(reader.read_offset_array(count)):
        reader.seek(offset)

        index = reader.read_uint16() if version == 4 else i

        tag_count = reader.read_uint16()
        tag_indexes = reader.read_uint16_array(tag_count)

        name = reader.read_str_variable_encoding()
        group_list.append(LMS_TagGroup(name, index, tag_indexes))

    return group_list



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.FileIO.Stream import FileReader
from LMS.Project.Definitions.Tag import LMS_TagParamInfo


def read_tgp2(reader: FileReader) -> list[LMS_TagParamInfo]:
    parameter_info = []

    count = reader.read_uint32()
    for offset in reader.read_offset_array(count):
        reader.seek(offset)

        info_type = LMS_DataType(reader.read_uint8())

        if info_type is not LMS_DataType.LIST:
            name = reader.read_str_variable_encoding()
            parameter_info.append(LMS_TagParamInfo(name, datatype=info_type))
            continue

        reader.skip(1)
        list_count = reader.read_uint16()
        list_indexes = reader.read_uint16_array(list_count)
        name = reader.read_str_variable_encoding()
        parameter_info.append(LMS_TagParamInfo(name, list_indexes, LMS_DataType.LIST))

    return parameter_info



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.FileIO.Stream import FileReader, FileWriter
from LMS.Message.Definitions.Field.LMS_Field import LMS_Field
from LMS.Message.Definitions.Field.LMS_FieldMap import LMS_FieldMap
from LMS.Message.Definitions.Field.Stream import read_field, write_field
from LMS.Message.Tag.LMS_Tag import LMS_DecodedTag, LMS_EncodedTag, LMS_TagBase
from LMS.Message.Tag.LMS_TagExceptions import (LMS_TagReadingError,
                                               LMS_TagWritingException)
from LMS.Message.Tag.System_Definitions import get_system_tag
from LMS.TitleConfig.Definitions.Tags import TagConfig, TagDefinition


def read_tag(
    reader: FileReader,
    param_size: int,
    group_index: int,
    tag_index: int,
    config: TagConfig | None,
) -> LMS_EncodedTag | LMS_DecodedTag:
    end = reader.tell() + param_size

    # System tags of group 0 are always defined, ensure they are read like any other decoded tag by default
    if config is None and group_index > 0:
        parameters = _read_encoded_parameters(reader, param_size)

        if parameters is None:
            return LMS_EncodedTag(group_index, tag_index)

        return LMS_EncodedTag(group_index, tag_index, parameters)

    if group_index == 0:
        definition = get_system_tag(tag_index)
    else:
        definition = config.get_definition_by_indexes(group_index, tag_index)

    # If the parameters were omitted from the definition but the tag still has defined parameters, add the decoded
    # names but read the tag as encoded. This is to account for encoded tags that group have tag names attatched.
    # i.e [Group:Tag 00-00-00-FF]
    if definition.parameters is None and param_size > 0:
        parameters = _read_encoded_parameters(reader, param_size)
        return LMS_EncodedTag(
            group_index,
            tag_index,
            parameters,
            definition.group_name,
            definition.tag_name,
        )
    else:
        parameters = _read_decoded_parameters(reader, definition)

    reader.seek(end)
    return LMS_DecodedTag(
        group_index, tag_index, definition.group_name, definition.tag_name, parameters
    )


def write_tag(writer: FileWriter, tag: LMS_TagBase) -> None:
    tag_indicator = b"\x0e" + (b"\x00" * (writer.encoding.width - 1))

    if writer.big_endian:
        tag_indicator = tag_indicator[::-1]

    writer.write_bytes(tag_indicator)
    writer.write_uint16(tag.group_index)
    writer.write_uint16(tag.tag_index)

    if tag.parameters is None:
        writer.write_uint16(0)
        return

    if isinstance(tag, LMS_EncodedTag):
        _write_encoded_parameters(writer, tag.parameters)
    else:
        _write_decoded_parameters(writer, tag)

    return


# --
def _read_encoded_parameters(reader: FileReader, param_size: int) -> list[str] | None:
    if param_size == 0:
        return

    hex_parameters = reader.read_bytes(param_size).hex().upper()
    encoded_parameters = [
        hex_parameters[i : i + 2] for i in range(0, len(hex_parameters), 2)
    ]
    return encoded_parameters


def _write_encoded_parameters(writer: FileWriter, parameters: list[str] | None) -> None:
    writer.write_uint16(len(parameters))
    for param in parameters:
        writer.write_bytes(bytes.fromhex(param))


def _read_decoded_parameters(
    reader: FileReader, definition: TagDefinition | None
) -> LMS_FieldMap | None:
    if definition.parameters is None:
        return

    parameters = {}
    for param in definition.parameters:
        param_offset = reader.tell()
        try:
            if param.datatype is LMS_DataType.STRING:
                value = LMS_Field(reader.read_len_string_variable_encoding(), param)
            else:
                value = read_field(reader, param)
        # There could be multiple errors related to reading, share the extra info but display the original exception
        except Exception as e:
            raise LMS_TagReadingError(
                f"An error occured reading tag '[{definition.group_name}:{definition.tag_name}]', parameter '{param.name}' at offset {param_offset}"
            ) from e

        parameters[param.name] = value
    return parameters


def _write_decoded_parameters(writer: FileWriter, tag: LMS_DecodedTag) -> None:
    param_size = 0

    # Tags are padded by a 0xCD byte if the size is not aligned to the encoding
    # This can occur before a string parameter, or at the end of the tag.
    # If a string parameter exists, then the padding will always be at the first string instance
    # first_string is a flag so that the first string can be padded correctly.
    first_string = True

    for field in tag.parameters.values():
        if field.datatype is LMS_DataType.STRING:
            param_size += 2 + len(field.value) * writer.encoding.width
        else:
            param_size += field.datatype.stream_size

    if needs_padding := param_size % 2 == 1:
        param_size += 1

    writer.write_uint16(param_size)
    for field in tag.parameters.values():
        try:
            if field.datatype is LMS_DataType.STRING:

                if first_string and needs_padding:
                    writer.write_bytes(b"\xcd")
                    first_string, needs_padding = False, False

                writer.write_len_variable_encoding_string(field.value)
            else:
                write_field(writer, field)
        except Exception as e:
            raise LMS_TagWritingException(
                f"An error occured writing tag '{tag}', parameter '{field.name}' at offset {writer.tell()}!"
            ) from e

    if needs_padding:
        writer.write_bytes(b"\xcd")



from __future__ import annotations

from LMS.Common.LMS_DataType import LMS_DataType


class LMS_TagGroup:
    def __init__(
        self, name: str = None, index: int = None, tag_indexes: list[int] = None
    ):
        self.name = name

        self.tag_definitions: list[LMS_TagInfo] = []

        # The index may be literally set or from the given from the stream
        # The index from the stream does not always align with the total amount of groups
        # Reason why is unknown (acts as an ID of some sort?)
        self.index = index

        self._tag_indexes = tag_indexes

    @property
    def tag_indexes(self) -> list[int]:
        return self._tag_indexes


class LMS_TagInfo:
    def __init__(self, name: str = None, param_indexes: list[int] = None):
        self.name = name
        self.param_info: list[LMS_TagParamInfo] = []
        self._parameter_indexes = param_indexes

    @property
    def parameter_indexes(self) -> list[int]:
        return self._parameter_indexes


class LMS_TagParamInfo:

    def __init__(
        self,
        name: str = None,
        list_indexes: list[int] = None,
        datatype: LMS_DataType = None,
        list_items: list[str] = None,
    ):
        self.name = name
        self.list_items = list_items or []

        self._list_indexes = [] if list_indexes is None else list_indexes
        self._type = datatype

    @property
    def datatype(self) -> LMS_DataType:
        return self._type

    @property
    def list_indexes(self) -> list[int]:
        return self._list_indexes



class LMS_Error(Exception):
    """Exception for non-specifc errors"""

    pass


class LMS_UnexpectedMagicError(Exception):
    """Exception for wrong magic of a LMS file"""

    pass


class LMS_MisalignedSizeError(Exception):
    """Exception for when size is not aligned."""

    pass



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.FileIO.Stream import FileReader, FileWriter
from LMS.Message.Definitions.Field.LMS_Field import LMS_Field
from LMS.TitleConfig.Definitions.Value import ValueDefinition


def read_field(reader: FileReader, definition: ValueDefinition) -> LMS_Field:
    # String is excluded as their reading varies between tags and attributes
    match definition.datatype:
        case LMS_DataType.UINT8:
            value = reader.read_uint8()
        case LMS_DataType.UINT16:
            value = reader.read_uint16()
        case LMS_DataType.UINT32:
            value = reader.read_uint32()
        case LMS_DataType.INT8:
            value = reader.read_int8()
        case LMS_DataType.INT16:
            value = reader.read_int16()
        case LMS_DataType.INT32:
            value = reader.read_int32()
        case LMS_DataType.FLOAT32:
            value = reader.read_float32()
        case LMS_DataType.LIST:
            index = reader.read_uint8()
            value = definition.list_items[index]
        case LMS_DataType.BOOL:
            value = bool(reader.read_uint8())
        case LMS_DataType.BYTE:
            value = reader.read_bytes(1)

    return LMS_Field(value, definition)


def write_field(writer: FileWriter, field: LMS_Field) -> None:
    match field.datatype:
        case LMS_DataType.UINT8:
            writer.write_uint8(field.value)
        case LMS_DataType.INT8:
            writer.write_int8(field.value)
        case LMS_DataType.UINT16:
            writer.write_uint16(field.value)
        case LMS_DataType.INT16:
            writer.write_int16(field.value)
        case LMS_DataType.UINT32:
            writer.write_uint32(field.value)
        case LMS_DataType.INT32:
            writer.write_int32(field.value)
        case LMS_DataType.LIST:
            writer.write_uint8(field.list_items.index(field.value))
        case LMS_DataType.BOOL:
            writer.write_uint8(bool(field.value))
        case LMS_DataType.BYTE:
            writer.write_bytes(field.value)



class LMS_Color:

    def __init__(
        self, red: int = None, green: int = None, blue: int = None, alpha: int = None
    ):
        self.name: str = None
        self.red = red
        self.green = green
        self.blue = blue
        self.alpha = alpha



# Regex formats for control tags

# Regular format for any tag
TAG_FORMAT = r"(\[[^\]]+\])"

DECODED_FORMAT = r"\[(\w+):(\w+)(.*?)\]"
ENCODED_FORMAT = r"\[([A-Za-z0-9]+):([A-Za-z0-9]+)\s?([0-9A-Fa-f\-]*)?\]"

# Key value pair regex for decoded tags
PARAMETER_FORMAT = r'(\w+)="([^"]*)"'



from typing import Any, Callable, Generator

from LMS.FileIO.Stream import FileReader, FileWriter


def read_section_data(
    reader: FileReader, section_count: int
) -> Generator[tuple[str, int], Any, None]:
    reader.seek(0x20)
    for _ in range(section_count):
        magic = reader.read_string_len(4)
        size = reader.read_uint32()

        # Skip to start of data
        reader.skip(8)
        end = reader.tell() + size

        yield (magic, size)

        # Seek past the AB padding on next iteration
        reader.seek(end)
        reader.align(16)


def write_section(
    writer: FileWriter,
    magic: str,
    section_call: Callable,
    data: list[Any],
    *write_parameters,
) -> None:
    writer.write_string(magic)
    size_offset = writer.tell()

    writer.write_uint32(0)
    writer.write_bytes(b"\x00" * 8)
    data_start = writer.tell()

    section_call(writer, data, *write_parameters)

    _write_end_data(writer, data_start, size_offset)


def write_unsupported_section(writer: FileWriter, magic: str, data: bytes) -> None:
    writer.write_string(magic)
    size_offset = writer.tell()
    writer.write_uint32(0)
    writer.write_bytes(b"\x00" * 8)
    data_start = writer.tell()
    writer.write_bytes(data)
    _write_end_data(writer, data_start, size_offset)


def _write_end_data(writer: FileWriter, data_start: int, size_offset: int) -> None:
    end = writer.tell()
    size = end - data_start
    writer.seek(size_offset)
    writer.write_uint32(size)
    writer.seek(end)
    writer.write_alignment(b"\xAB", 16)



from LMS.FileIO.Stream import FileReader, FileWriter


def read_tsy1(reader: FileReader, message_count: int) -> list[int]:
    style_indexes = []
    for _ in range(message_count):
        style_indexes.append(reader.read_uint32())
    return style_indexes


def write_tsy1(writer: FileWriter, style_indexes: list[int]) -> None:
    for i in style_indexes:
        writer.write_uint32(i)



from LMS.FileIO.Encoding import FileEncoding
from LMS.FileIO.Stream import FileReader, FileWriter
from LMS.Message.Definitions.LMS_MessageText import LMS_MessageText
from LMS.Message.Tag.LMS_Tag import LMS_TagBase
from LMS.Message.Tag.Stream import read_tag, write_tag
from LMS.TitleConfig.Definitions.Tags import TagConfig


def read_txt2(
    reader: FileReader, encoding: FileEncoding, config: TagConfig
) -> list[LMS_MessageText]:
    text_list = []
    count = reader.read_uint32()

    encoding_format = encoding.to_string_format(reader.big_endian)
    tag_indicator = b"\x0e" + (b"\x00" * (reader.encoding.width - 1))

    if reader.big_endian:
        tag_indicator = tag_indicator[::-1]

    for offset in reader.read_offset_array(count):
        reader.seek(offset)

        text, parts = b"", []
        while (data := reader.read_bytes(encoding.width)) != encoding.terminator:
            if data == tag_indicator:
                # Add all the text before the control tag
                parts.append(text.decode(encoding_format))

                # Read the tag data
                group_index = reader.read_uint16()
                tag_index = reader.read_uint16()
                param_size = reader.read_uint16()

                tag = read_tag(reader, param_size, group_index, tag_index, config)

                parts.append(tag)
                text = b""
            # TODO: Add support for closing tags
            elif data == b"\x0f":
                raise NotImplementedError(
                    "MSBT files with closing tags are unsupported at this time!"
                )
            else:
                text += data

        # Add the remaining text in case there were no control tags
        if text:
            parts.append(text.decode(encoding_format))

        message = LMS_MessageText(parts, config)
        text_list.append(message)

    return text_list


def write_txt2(writer: FileWriter, messages: list[LMS_MessageText]) -> None:
    start = writer.tell()
    writer.write_uint32(len(messages))

    offset = 4 + 4 * len(messages)
    for message in messages:
        writer.write_uint32(offset)
        next = writer.tell()

        writer.seek(start + offset)
        text_start = writer.tell()
        for part in message:
            if isinstance(part, LMS_TagBase):
                write_tag(writer, part)
            else:
                writer.write_variable_encoding_string(part, False)

        writer.write_bytes(writer.encoding.terminator)
        offset += writer.tell() - text_start
        writer.seek(next)

    # Seek to the end of the text to ensure end data can be written properly
    writer.seek(offset + start)



from enum import Enum
from typing import Literal, Type


class LMS_DataType(Enum):

    UINT8 = 0
    UINT16 = 1
    UINT32 = 2

    INT8 = 3
    INT16 = 4
    INT32 = 5

    FLOAT32 = 6

    # Unknown 16 bit type (value of 6) has yet to be documented
    ...

    STRING = 8
    LIST = 9

    # Interface types
    BOOL = "bool"
    BYTE = "byte"

    def to_string(self) -> str:
        return self._name_.lower()

    @property
    def signed(self) -> bool:
        """Property for if the type is signed or not."""
        if self not in [
            LMS_DataType.STRING,
            LMS_DataType.LIST,
            LMS_DataType.BOOL,
            LMS_DataType.BYTE,
        ]:
            return self in [LMS_DataType.INT8, LMS_DataType.INT16, LMS_DataType.INT32]

        raise TypeError(f"Signed is not a valid property for '{self.to_string()}'!")

    @property
    def builtin_type(self) -> Type[int | str | float | bool | bytes]:
        """The enum as the builtin python type."""
        return {
            LMS_DataType.UINT8: int,
            LMS_DataType.UINT16: int,
            LMS_DataType.UINT32: int,
            LMS_DataType.INT8: int,
            LMS_DataType.INT16: int,
            LMS_DataType.INT32: int,
            LMS_DataType.FLOAT32: float,
            LMS_DataType.STRING: str,
            LMS_DataType.LIST: str,
            LMS_DataType.BOOL: bool,
            LMS_DataType.BYTE: bytes,
        }[self]

    @property
    def stream_size(self) -> Literal[1, 2, 4]:
        """The size the datatype takes up in a strema."""
        return {
            LMS_DataType.UINT8: 1,
            LMS_DataType.UINT16: 2,
            LMS_DataType.UINT32: 4,
            LMS_DataType.INT8: 1,
            LMS_DataType.INT16: 2,
            LMS_DataType.INT32: 4,
            LMS_DataType.FLOAT32: 4,
            LMS_DataType.LIST: 1,
            LMS_DataType.BOOL: 1,
            LMS_DataType.BYTE: 1,
        }[self]

    @classmethod
    def from_string(cls, string: str):
        """Creates a LMS_Datatype enum value from it's string representation"""
        string = string.upper()
        if string in cls.__members__:
            return cls[string]
        else:
            raise ValueError(f"Unknown value of '{string}' was provided!")



from LMS.Common.LMS_DataType import LMS_DataType


class LMS_AttributeInfo:
    def __init__(self, datatype: LMS_DataType, offset: int, list_index: int = None):
        super().__init__()

        self.name: str = None

        # Set later on from the ALI2 block
        self.list_items: list[str] = []

        self._datatype = datatype
        self._offset = offset
        self._list_index = list_index

    @property
    def datatype(self) -> LMS_DataType:
        return self._datatype

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def list_index(self) -> int:
        return self._list_index



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.FileIO.Stream import FileReader
from LMS.Project.Definitions.Attribute import LMS_AttributeInfo


def read_ati2(reader: FileReader) -> list[LMS_AttributeInfo]:
    info_list = []
    count = reader.read_uint32()
    for _ in range(count):
        datatype = LMS_DataType(reader.read_uint8())
        reader.skip(1)
        list_index = reader.read_uint16()
        offset = reader.read_uint32()

        info_list.append(LMS_AttributeInfo(datatype, offset, list_index))

    return info_list



from dataclasses import dataclass, field

from LMS.Common.LMS_DataType import LMS_DataType


@dataclass(frozen=True)
class ValueDefinition:
    name: str
    description: str
    datatype: LMS_DataType
    list_items: dict[int, str] = field(default=None)

    @classmethod
    def from_dict(cls, data: dict):
        name, description = data["name"], data["description"]
        datatype = LMS_DataType.from_string(data["datatype"])
        list_items = data.get("list_items")
        return cls(name, description, datatype, list_items)



from LMS.Common import LMS_Exceptions
from LMS.Common.LMS_FileInfo import LMS_FileInfo
from LMS.FileIO.Encoding import FileEncoding
from LMS.FileIO.Stream import FileReader, FileWriter


def read_file_info(reader: FileReader, expected_magic: str) -> LMS_FileInfo:
    magic = reader.read_string_len(8)

    if magic != expected_magic:
        raise LMS_Exceptions.LMS_UnexpectedMagicError(
            f"Invalid magic!' Expected {expected_magic}', got '{magic}'."
        )

    big_endian = reader.read_bytes(2) == b"\xfe\xff"
    reader.big_endian = big_endian

    reader.skip(2)

    encoding = FileEncoding(reader.read_uint8())
    reader.encoding = encoding

    version = reader.read_uint8()
    section_count = reader.read_uint16()

    reader.skip(2)
    file_size = reader.read_uint32()

    reader.seek(0, 2)
    if file_size != reader.tell():
        raise LMS_Exceptions.LMS_MisalignedSizeError(f"Filesize is misaligned!")

    # Seek to the start of data
    reader.seek(0x20)

    return LMS_FileInfo(
        big_endian,
        encoding,
        version,
        section_count,
    )


def write_file_info(writer: FileWriter, magic: str, file_info: LMS_FileInfo) -> None:
    """Writes the file info to a stream.

    :param writer: a Writer object.
    :param file_info: the file_info object."""
    writer.big_endian = file_info.big_endian
    writer.encoding = file_info.encoding

    writer.write_string(magic)
    writer.write_bytes(b"\xff\xfe" if not file_info.big_endian else b"\xfe\xff")
    writer.write_bytes(b"\x00\x00")

    writer.write_uint8(file_info.encoding.value)
    writer.write_uint8(file_info.version)
    writer.write_uint16(file_info.section_count)

    # Padding
    writer.write_bytes(b"\x00\x00")
    writer.write_bytes(b"\x00" * 4)
    writer.write_bytes(b"\x00" * 10)
    writer.seek(0x20)



from LMS.Message.Definitions.Field.LMS_Field import LMS_Field

# Typehint that represents a map of strings to field instances.
# Utilized as a means of storing attributes, or parameters in decoded tags mapped to their string names.
type LMS_FieldMap = dict[str, LMS_Field]



import re
from abc import ABC, abstractmethod

from LMS.Message.Definitions.Field.LMS_Field import LMS_Field, cast_value
from LMS.Message.Definitions.Field.LMS_FieldMap import LMS_FieldMap
from LMS.Message.Tag.LMS_TagExceptions import LMS_InvalidTagFormatError
from LMS.Message.Tag.Tag_Formats import (DECODED_FORMAT, ENCODED_FORMAT,
                                         PARAMETER_FORMAT)
from LMS.TitleConfig.Definitions.Tags import TagConfig


class LMS_TagBase(ABC):
    def __init__(
        self,
        group_index: int,
        tag_index: int,
        parameters: LMS_FieldMap | list[str] | None = None,
        group_name: str = None,
        tag_name: str = None,
    ):
        self._group_index = group_index
        self._tag_index = tag_index
        self._parameters = parameters or {}
        self._group_name, self._tag_name = group_name, tag_name

    @property
    def group_index(self) -> int:
        """The index of the group."""
        return self._group_index

    @property
    def tag_index(self) -> int:
        """The index of the tag in it's group."""
        return self._tag_index

    @property
    def group_name(self) -> str:
        """The name of the tag group."""
        return self._group_name

    @property
    def tag_name(self) -> str:
        """The name of the tag in the tag group."""
        return self._tag_name

    @property
    def parameters(self) -> list[str] | LMS_FieldMap:
        """The parameters of the tag."""
        return self._parameters

    @abstractmethod
    def to_text(self) -> str:
        """Converts the tag to its string representation."""
        pass

    @classmethod
    @abstractmethod
    def from_string(self, string: str, config: TagConfig | None = None) -> str:
        pass


class LMS_EncodedTag(LMS_TagBase):
    """A class that represents an encoded tag.

    Example encoded tags:
        - `[0:3 00-00-00-FF]`
        - `[0:4]`
        - `[1:0 01-00-00-CD]`"""

    def __init__(
        self,
        group_index: int,
        tag_index: int,
        parameters: list[str] = None,
        group_name: str = None,
        tag_name: str = None,
    ):
        super().__init__(group_index, tag_index, parameters)
        self._group_name = group_name
        self._tag_name = tag_name

    def to_text(self) -> str:
        if self._group_name is not None and self._tag_name is not None:
            # Determine what to display based on if the names are provided
            group = self._group_name or self._group_index
            tag = self._tag_name or self._tag_index
        else:
            group, tag = self._group_index, self._tag_index

        if not self._parameters:
            return f"[{group}:{tag}]"

        parameters = "-".join(self.parameters)
        return f"[{group}:{tag} {parameters}]"

    @classmethod
    def from_string(cls, string: str, config: TagConfig | None = None):
        if match := re.match(ENCODED_FORMAT, string):
            group, tag = match.group(1), match.group(2)
            parameters = match.group(3).split("-")

            if group.isdigit() and tag.isdigit():
                return cls(int(group), int(tag), parameters)

            tag_definition = config.get_definition_by_names(group, tag)
            return cls(
                group,
                tag,
                parameters,
                tag_definition.group_name,
                tag_definition.tag_name,
            )
        else:
            raise LMS_InvalidTagFormatError(
                f"Invalid encoded tag format detected for tag'{string}'"
            )


class LMS_DecodedTag(LMS_TagBase):
    """A class that represents a decoded tag.

    Example encoded tags:
        - `[System:Color r="0" g="255" b="255" a="255"]`
        - `[System:Pagebreak]`
        - `[Mii:Nickname buffer="1" type="Text" conversion="None"]`"""

    def __init__(
        self,
        group_index: int,
        tag_index: int,
        group_name: str,
        tag_name: str,
        parameters: LMS_FieldMap = None,
    ):
        super().__init__(group_index, tag_index, parameters, group_name, tag_name)

    def to_text(self) -> str:
        if not self._parameters:
            return f"[{self.group_name}:{self.tag_name}]"

        parameters = " ".join(
            f'{key}="{param._value}"' for key, param in self._parameters.items()
        )
        return f"[{self.group_name}:{self.tag_name} {parameters}]"

    @classmethod
    def from_string(cls, tag: str, config: TagConfig):
        if match := re.match(DECODED_FORMAT, tag):
            group_name, tag_name = match.group(1), match.group(2)
            parameters = dict(re.findall(PARAMETER_FORMAT, tag))

            tag_definition = config.get_definition_by_names(group_name, tag_name)
            for definition in tag_definition.parameters:
                casted_value = cast_value(
                    parameters[definition.name], definition.datatype
                )
                parameters[definition.name] = LMS_Field(casted_value, definition)

            return cls(
                tag_definition.group_index,
                tag_definition.tag_index,
                group_name,
                tag_name,
                parameters,
            )
        else:
            raise LMS_InvalidTagFormatError(
                f"Invalid decoded tag format detected for tag '{tag}'"
            )



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.TitleConfig.Definitions.Tags import TagDefinition
from LMS.TitleConfig.Definitions.Value import ValueDefinition


def get_system_tag(group: int | str) -> TagDefinition:
    return SYSTEM_INT_MAP[group] if isinstance(group, int) else SYSTEM_STR_MAP[group]


RUBY_TAG = TagDefinition(
    group_name="System",
    group_index=0,
    tag_name="Ruby",
    tag_index=0,
    description="Displays a RUBY character.",
    parameters=[
        ValueDefinition(
            name="rt", description="The ruby character.", datatype=LMS_DataType.STRING
        )
    ],
)

FONT_TAG = TagDefinition(
    group_name="System",
    group_index=0,
    tag_name="Font",
    tag_index=1,
    description="Sets the font file of text.",
    parameters=[
        ValueDefinition(
            name="face", description="The font face name.", datatype=LMS_DataType.STRING
        )
    ],
)

SIZE_TAG = TagDefinition(
    group_name="System",
    group_index=0,
    tag_name="Size",
    tag_index=2,
    description="Alters the size of text.",
    parameters=[
        ValueDefinition(
            name="percent",
            description="The new size as a percent.",
            datatype=LMS_DataType.UINT16,
        )
    ],
)
COLOR_TAG = TagDefinition(
    group_name="System",
    group_index=0,
    tag_name="Color",
    tag_index=3,
    description="Sets the color of text in a message in RGBA form.",
    parameters=[
        ValueDefinition(
            name="r", description="The red color value.", datatype=LMS_DataType.UINT8
        ),
        ValueDefinition(
            name="g", description="The green color value.", datatype=LMS_DataType.UINT8
        ),
        ValueDefinition(
            name="b", description="The blue color value.", datatype=LMS_DataType.UINT8
        ),
        ValueDefinition(
            name="a",
            description="The alpha transprency value.",
            datatype=LMS_DataType.UINT8,
        ),
    ],
)

PAGEBREAK_TAG = TagDefinition(
    group_name="System",
    group_index=0,
    tag_name="PageBreak",
    tag_index=4,
    description="Displays text in the same label on different pages.",
)



from LMS.Common.LMS_DataType import LMS_DataType
from LMS.FileIO.Stream import FileReader
from LMS.Project.Definitions.Tag import LMS_TagInfo


def read_tag2(reader: FileReader) -> list[LMS_TagInfo]:
    info_list = []

    count = reader.read_uint32()
    for offset in reader.read_offset_array(count):
        reader.seek(offset)

        param_count = reader.read_uint16()
        param_indexes = reader.read_uint16_array(param_count)
        name = reader.read_str_variable_encoding()

        info_list.append(LMS_TagInfo(name, param_indexes))

    return info_list



from enum import Enum
from typing import Literal


class FileEncoding(Enum):
    """An enum that represents a file encoding."""

    UTF8 = 0
    UTF16 = 1
    UTF32 = 2

    def to_string_format(
        self, big_endian: bool = False
    ) -> Literal["UTF-8", "UTF-16-LE", "UTF-16-BE", "UTF-32-LE", "UTF-32-BE"]:
        """Returns the encoding in the string format"""
        match self:
            case FileEncoding.UTF8:
                return "UTF-8"
            case FileEncoding.UTF16:
                if big_endian:
                    return "UTF-16-BE"
                return "UTF-16-LE"
            case FileEncoding.UTF32:
                if big_endian:
                    return "UTF-32-BE"
                return "UTF-32-LE"

    @property
    def width(self) -> Literal[1, 2, 4]:
        """The width of a character in a stream."""
        match self:
            case FileEncoding.UTF8:
                return 1
            case FileEncoding.UTF16:
                return 2
            case FileEncoding.UTF32:
                return 4

    @property
    def terminator(self) -> Literal[b"\x00", b"\x00\x00", b"\x00\x00\x00\x00"]:
        """The terminator of the string in a stream."""
        return b"\x00" * self.width
